/*
    CharityWorld: Agent.m
    Copyright (C) 2003  Macaulay Institute

    This file is part of CharityWorld, a simple agent-based model to
    demonstrate the effects of errors in floating point arithmetic on
    emergent outcomes.

    CharityWorld is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CharityWorld is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/

#import "Agent.h"
#import "ModelSwarm.h"
#import "MergeSorter.h"
#import "Shuffler.h"
#import "Verbosity.h"
#import "DoubleSimple.h"
#import <random.h>
#import <math.h>
#import <ieeefp.h>
#import <float.h>

@implementation Agent

+create: (id <Zone>)z {
  Agent *obj = [super create: z];

  obj->nbrs = nil;
  obj->donatedNbrs = nil;
  obj->beg = nil;
  obj->generosity = nil;
  obj->wealth = nil;
  obj->localWealth = nil;
  obj->minimum = nil;
  obj->maximum = nil;
  obj->numberOfNbrs = 0;
  return obj;
}

+create: (id <Zone>)z model: (ModelSwarm *)m X: (int)ax Y: (int)ay {
  Agent *obj;

  obj = [super create: z];
  obj->model = m;
  obj->wealth = [DoubleSimple new];
  obj->localWealth = [DoubleSimple new];
  obj->beg = [DoubleSimple new];
  obj->generosity = [DoubleSimple new];
  obj->minimum = [DoubleSimple new];
  obj->maximum = [DoubleSimple new];
  obj->nbrs = [List create: z];
  obj->donatedNbrs = [List create: z];
  obj->numberOfNbrs = 0;
  obj->x = ax;
  obj->y = ay;
  obj->waiting = NO;

  return obj;
}

-(int)getX {
  return x;
}

-(int)getY {
  return y;
}

/*

addNeighbour:

This method is called from the model when the agent is being created,
and ensures that the neighbouring agents are in an arbitrary order.

*/

-addNeighbour: (Agent *)a {
  
  if(![nbrs contains: a]) {
    [nbrs addLast: a];
    numberOfNbrs++;
  }
  [Shuffler shuffleList: nbrs];

  return self;
}

/*

addDonatedNeighbour:

This method is called from the Bank whenever a traced transfer is made from
one agent to another.

*/

-addDonatedNeighbour: (Agent *)a {
  if(![donatedNbrs contains: a]) {
    [donatedNbrs addLast: a];
  }

  return self;
}

/*

clearDonatedNeighbours

This method may be called from the ObserverSwarm after displaying a trace.
N.B. This means donatedNbrs should NOT be used for checking donations.

*/

-clearDonatedNeighbours {
  [donatedNbrs removeAll];
  return self;
}

/*

These methods set the amount the agent begs for when begging and the
amount the agent offers when donating -- each is called from the
modelSwarm.

*/

-setBeg: (DoubleSimple *)b {
  [beg set: b];
  return self;
}

-setGenerosity: (DoubleSimple *)g {
  [generosity set: g];
  return self;
}

/*

computeLocalWealth

Computes the local average wealth of the agent -- the average of the
wealth of the agent and its neighbours.

*/

-(void)computeLocalWealth {
  id <Index> ix;
  Agent *nbr;

  [localWealth set: wealth];
  [minimum set: [maximum set: wealth]];
  if([Verbosity showAgentAverage]) {
    printf("Examining neighbouring wealth. Mine (%d, %d) = ", x + 1, y + 1);
    [wealth writeAccurately: stdout];
    printf("\n\tTotal = ");
    [localWealth writeAccurately: stdout];
    printf(", minimum = ");
    [minimum writeAccurately: stdout];
    printf(", maximum = ");
    [maximum writeAccurately: stdout];
    printf("\n");
  }
  for(ix = [nbrs begin: scratchZone], nbr = (Agent *)[ix next];
      [ix getLoc] == Member;
      nbr = (Agent *)[ix next]) {
    DoubleSimple *nbrWealth = [nbr getWealth];
    [localWealth add: nbrWealth];
    [minimum set: [minimum gt: nbrWealth] ? nbrWealth : minimum];
    [maximum set: [maximum lt: nbrWealth] ? nbrWealth : maximum];
    if([Verbosity showAgentAverage]) {
      printf("\tNeighbouring wealth: ");
      [nbrWealth writeAccurately: stdout];
      printf("\n\tTotal = ");
      [localWealth writeAccurately: stdout];
      printf(", minimum = ");
      [minimum writeAccurately: stdout];
      printf(", maximum = ");
      [maximum writeAccurately: stdout];
      printf("\n");
    }
    [nbrWealth free];
  }
  [ix drop];
  if([Verbosity showAgentAverage]) {
    printf("\tLocal Wealth = ");
    [localWealth writeAccurately: stdout];
    printf("\n");
  }
}

/*

The following methods need to be defined by any subclassing agent, but
are not defined here.

*/

-(BOOL)beggarCondition {
  printf("WARNING: Abstract Agent asked for beggar condition!\n");
  return NO;
}

-(BOOL)donorCondition {
  printf("WARNING: Abstract Agent asked for donor condition!\n");
  return NO;
}

-(BOOL)respondingCondition: (Agent *)nbr {
  printf("WARNING: Abstract Agent asked for offering condition!\n");
  return NO;
}

-(BOOL)acceptingCondition: (Agent *)nbr {
  printf("WARNING: Abstract Agent asked for accepting condition!\n");
  return NO;
}

-(BOOL)offeringCondition: (Agent *)nbr {
  printf("WARNING: Abstract Agent asked for offering condition!\n");
  return NO;
}

-(BOOL)askingCondition: (Agent *)nbr {
  printf("WARNING: Abstract Agent asked for accepting condition!\n");
  return NO;
}

-(BOOL)happinessCondition {
  printf("WARNING: Abstract Agent asked for happiness condition!\n");
  return NO;
}

-step {
  printf("WARNING: Abstract Agent asked to step!\n");
  return self;
}

-ask { 
  printf("WARNING: Abstract Agent asked to ask!\n");
  return self;
}

-offer {
  printf("WARNING: Abstract Agent asked to offer!\n");
  return self;
}


-(void)accept: (DoubleSimple *)amount from: (Agent *)a {
  printf("WARNING: Abstract Agent asked to receive!\n");
}

-(void)respondTo: (Agent *)a askingFor: (DoubleSimple *)amount {
  printf("WARNING: Abstract Agent asked to respond!\n");
}

/*

Methods to add and remove wealth from the agent

*/

-(void)addWealth: (DoubleSimple *)amount {
  if([Verbosity showAgentChangingWealth]) {
    printf("(%d, %d) received ", x + 1, y + 1);
    [amount writeAccurately: stdout];
    printf(" of wealth\n");
  }
  [wealth add: amount];
}

-(void)subWealth: (DoubleSimple *)amount {
  if([Verbosity showAgentChangingWealth]) {
    printf("(%d, %d) lost ", x + 1, y + 1);
    [amount writeAccurately: stdout];
    printf(" of wealth\n");
  }
  [wealth sub: amount];
}

/*

There are two methods for rescheduling. The rescheduleCheck method
computes the local average wealth before rescheduling, and the
reschedule method calls the beggarCondition and donorCondition methods
to determine which list, if any, the agent should reschedule itself
to. An agent that has put itself on a list is said to be waiting until
it is asked to reschedule and does not put itself on any list.

*/

-(void)rescheduleCheck {
  [self computeLocalWealth];
  [self reschedule];
}

-(void)reschedule {
  waiting = YES;
  if([self beggarCondition]) {
    if([Verbosity showAgentRescheduling]) {
      printf("(%d, %d) queued to beggar list\n", x + 1, y + 1);
    }
    [model queueBeggar: self];
  }
  else if([self donorCondition]) {
    if([Verbosity showAgentRescheduling]) {
      printf("(%d, %d) queued to donor list\n", x + 1, y + 1);
    }
    [model queueDonor: self];
  }
  else {
    if([Verbosity showAgentRescheduling]) {
      printf("(%d, %d) not queued to any list\n", x + 1, y + 1);
    }
    waiting = NO;
  }
}

-(BOOL)isWaiting {
  return waiting;
}

-(BOOL)donatedTo: (Agent *)a {
  return [donatedNbrs contains: a] ? YES : NO;
}

-(DoubleSimple *)getWealth {
  return [wealth copy];
}

-(double)getWealthForDisplayOnly {
  return [wealth get];
}

-(double)getMinimumForDisplayOnly {
  return [minimum get];
}

-(double)getMaximumForDisplayOnly {
  return [maximum get];
}

-(double)getBegForDisplayOnly {
  return [beg get];
}

-(double)getGenerosityForDisplayOnly {
  return [generosity get];
}

/*

The next two methods are called from the observerSwarm raster objects
when the agent is asked to draw itself on the raster. drawWealthOn:
draws the agent on the raster with a colour indicating the absolute
wealth in terms of this simulation -- with minimum equal to the
jackpot divided by the number of agents and the maximum equal to the
jackpot (in fact, it is the jackpot minus the ticket price...)

*/

-drawWealthOn: r {
  int color;
  double w = [wealth get];
  double maxw = [model getMaxWealthForDisplayOnly];
  double minw = [model getMinWealthForDisplayOnly];
  double neutralw = [model getInitialWealthForDisplayOnly];

  [self computeLocalWealth];
  if([wealth clt: minw]) {
    color = 0;
  }
  else if([wealth cgt: maxw]) {
    color = 128;
  }
  else if([wealth clt: neutralw]) {
    double bigmin, linear;
    if(fabs(neutralw - minw) < fabs(maxw - neutralw)) {
      bigmin = minw;
    }
    else if(fabs(neutralw - minw) > fabs(maxw - neutralw)) {
      bigmin = neutralw - fabs(maxw - neutralw);
    }
    else {
      bigmin = neutralw - (fabs(neutralw - minw) / 5.0);
    }
    if(w < bigmin) {
      linear = 1.0 - ((bigmin - w) / (bigmin - minw));
      color = 130 + (int)(31.0 * linear);
    }
    else {
      linear = 1.0 - ((neutralw - w) / (neutralw - bigmin));
      color = (int)(63.0 * linear);
    }
  }
  else if([wealth cgt: neutralw]) {
    double smallmax, linear;
    if(fabs(neutralw - minw) < fabs(maxw - neutralw)) {
      smallmax = neutralw + fabs(neutralw - minw);
    }
    else if(fabs(neutralw - minw) > fabs(maxw - neutralw)) {
      smallmax = maxw;
    }
    else {
      smallmax = neutralw + (fabs(maxw - neutralw) / 5.0);
    }
    if(w > smallmax) {
      linear = (w - smallmax) / (maxw - smallmax);
      color = 162 + (int)(31.0 * linear);
    }
    else {
      linear = (w - neutralw) / (smallmax - neutralw);
      color = 65 + (int)(63.0 * linear);
    }
  }
  else {
    color = 64;
  }
  [r drawPointX: x Y: y Color: color];
  return self;
}

-drawTracedDonationsOn: r {
  id ix;
  Agent *nbr;

  for(ix = [donatedNbrs begin: scratchZone], nbr = (Agent *)[ix next];
      [ix getLoc] == Member;
      nbr = (Agent *)[ix next]) {
    [r lineX0: x Y0: y X1: [nbr getX] Y1: [nbr getY] Width: 0 Color: 129];
  }
  [ix drop];
  return self;
}

-drawLocalWealthOn: r {
  int color;

  [self computeLocalWealth];
  if([maximum eq: minimum]) {
    color = 64;
  }
  else if([wealth lt: minimum]) {
    color = 0;
  }
  else if([wealth gt: maximum]) {
    color = 128;
  }
  else {
    color = (int)(127.0 * ([wealth get] - [minimum get])
		  / ([maximum get] - [minimum get]));
    if(color >= 64) color++;
  }
  [r drawPointX: x Y: y Color: color];
  return self;
}

-drawFloatingPointClassOn: r {
#ifndef __CYGWIN__
  switch(fpclass([wealth get])) {
  case FP_SNAN:
    [r drawPointX: x Y: y Color: 201];
    break;
  case FP_QNAN:
    [r drawPointX: x Y: y Color: 202];
    break;
  case FP_NINF:
    [r drawPointX: x Y: y Color: 203];
    break;
  case FP_PINF:
    [r drawPointX: x Y: y Color: 204];
    break;
  case FP_NDENORM:
    [r drawPointX: x Y: y Color: 205];
    break;
  case FP_PDENORM:
    [r drawPointX: x Y: y Color: 206];
    break;
  case FP_NZERO:
    [r drawPointX: x Y: y Color: 207];
    break;
  case FP_PZERO:
    [r drawPointX: x Y: y Color: 208];
    break;
  case FP_NNORM:
    [r drawPointX: x Y: y Color: 209];
    break;
  case FP_PNORM:
    [r drawPointX: x Y: y Color: 210];
    break;
  default:
    [r drawPointX: x Y: y Color: 200];
  }
#endif
  return self;
}

-drawLog2WealthOn: r {
  double w = [wealth get];
  int log2 = ilogb(w);

  if(log2 == -INT_MAX) {
    [r drawPointX: x Y: y Color: 200];
  }
  else if(log2 == INT_MAX) {
    [r drawPointX: x Y: y Color: 203];
  }
  else if(log2 <= DBL_MIN_EXP) {
    if(w >= 0.0) {
      [r drawPointX: x Y: y Color: 201];
    }
    else {
      [r drawPointX: x Y: y Color: 202];
    }
  }
  else if(w < 0.0) {
    if(log2 < 0) {
      [r drawPointX: x Y: y Color: ((49 * (-1 - log2)) / 1021) + 100];
    }
    else {
      [r drawPointX: x Y: y Color: ((49 * log2) / 1023) + 150];
    }
  }
  else if(w > 0.0) {
    if(log2 < 0) {
      [r drawPointX: x Y: y Color: ((49 * (-1 - log2)) / 1021) + 50];
    }
    else {
      [r drawPointX: x Y: y Color: ((49 * log2) / 1023)];
    }
  }
  else {
    [r drawPointX: x Y: y Color: 204];
  }
  return self;
}

-drawScheduleOn: r {
  id <List> donors = [[model getDonors] getList];
  id <List> beggars = [[model getBeggars] getList];

  if([donors contains: self]) {
    if([beggars contains: self]) {
      [r drawPointX: x Y: y Color: 129];
    }
    else {
      [r drawPointX: x Y: y Color: 193];
    }
  }
  else if([beggars contains: self]) {
    [r drawPointX: x Y: y Color: 130];
  }	
  else {
    [r drawPointX: x Y: y Color: 64];
  }
  return self;
}  

-printLocation{
  printf("(%d, %d)", x + 1, y + 1);
  return self;
}

-printWealth{
  [wealth writeAccurately: stdout];
  return self;
}

-(void)drop {
  if(nbrs != nil) [nbrs drop];
  if(wealth != nil) [wealth free];
  if(localWealth != nil) [localWealth free];
  if(minimum != nil) [minimum free];
  if(maximum != nil) [maximum free];
  if(beg != nil) [beg free];
  if(generosity != nil) [generosity free];
  [super drop];
}

@end
