/*
    CharityWorld: AllPaperAgents.h
    Copyright (C) 2003  Macaulay Institute

    This file is part of CharityWorld, a simple agent-based model to
    demonstrate the effects of errors in floating point arithmetic on
    emergent outcomes.

    CharityWorld is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CharityWorld is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/* 

This file contains declarations for all the agents used in the
paper. The following outlines the class hierarchy to show how the
relevant methods for each class of agent work. The following variables
act as shorthand for their corresponding quantities:

L - local wealth, computed by method computeLocalWealth
W - the wealth of the agent that is self
R - the wealth of a (potential) recipient agent of self
S - the sum of the wealth of the neighbours of self
N - the number of agents in self's neighbourhood
G - the generosity parameter supplied to the model

The following methods in the classes correspond to actions in the paper:

Supply-Awareness-Condition(A) :- donorCondition, shortened to A[] here.
Supply-Approach-Condition(A, B) :- offeringCondition: B, shortened to O[] here.
Amount to donate, g :- computed in offer, shortened to D[] here.

The SUA classes use "happiness" to determine awareness/approach, and
this is determined using the happinessCondition method, shortened to
H[] here.

Agent: L = S + W.
 |
 +-- SimpleAgent: A[ W * (N + 1) > L ];
      |           O[ R * (N + 1) < L ];
      |           D[ G ];
      |           H[ W * (N + 1) > L ].
      |
      +-- SimpleSupplyAgent: (disables non-supply features of SimpleAgent)
      |    |
      |    +-- LocalSharingSupplyAgent: O[ always ];
      |    |                            D[ G * (W - (S / N)) / (N + 1) ].
      |    |
      |    +-- HappySupplyAgent: O[ not H[ neighbour ] ].
      |
      +-- MeanSimpleAgent: L = (S + W) / (N + 1);
      |    |               A[ W > L ];
      |    |               O[ R < L ];
      |    |               H[ W > L ].
      |    |
      |    +-- MeanSimpleSupplyAgent: (disables non-supply features)
      |    |    |
      |    |    +-- MeanHappySupplyAgent: O[ not H[ neighbour ] ].
      |    |
      |    +-- ExtMeanSimpleAgent: L = S / N.
      |         |
      |         +-- ExtMeanSimpleSupplyAgent: (disables non-supply features)
      |              |
      |              +-- ExtMeanHappySupplyAgent: O[ not H[ neighbour ] ].
      |
      +-- ExtAvgSimpleAgent: L = S;
           |                 A[ W * N > L ];
	   |                 O[ R * N < L ];
	   |                 H[ W * N > L ].
           |
           +-- ExtAvgSimpleSupplyAgent: (disables non-supply features)
                |
                +-- ExtAvgHappySupplyAgent: O[ not H[ neighbour ] ].

*/

#import "AllSimpleSupplyAgents.h"
#import "AllLocalSharingSupplyAgents.h"
#import "AllHappySupplyAgents.h"

/*

SSA == Simple-Supply-Agent

L = S + W;
A[ W * (N + 1) > L ];
O[ R * (N + 1) < L ];
D[ G ];
H[ W * (N + 1) > L ].

*/

@interface SSA : SimpleSupplyAgent {
}

@end

/*

SSAMean == Simple-Supply-Agent, mean variant

L = (S + W) / (N + 1);
A[ W > L ];
O[ R < L ];
D[ G ];
H[ W > L ];

*/

@interface SSAMean : MeanSimpleSupplyAgent {
}

@end

/*

SSATotal == Simple-Supply-Agent, total variant (This is the same as SSA)

L = S + W;
A[ W * (N + 1) > L ];
O[ R * (N + 1) < L ];
D[ G ];
H[ W * (N + 1) > L ].

*/

@interface SSATotal : SimpleSupplyAgent {
}

@end

/*

PSA == Portion-Supply-Agent

L = S + W;
A[ W * (N + 1) > L ];
O[ always ];
D[ G * (W - (S / N)) / (N + 1) ];
H[ W * (N + 1) > L ].

*/

@interface PSA : LocalSharingSupplyAgent {
}

@end

/*

SUAInclusiveMean == Supply-Unhappy-Agent, inclusive mean variant

L = (S + W) / (N + 1);
A[ W > L ];
O[ not H[ neighbour ] ];
D[ G ];
H[ W > L ].

*/

@interface SUAInclusiveMean : MeanHappySupplyAgent {
}

@end

/*

SUAInclusiveTotal == Supply-Unhappy-Agent, inclusive total variant

L = S + W;
A[ W * (N + 1) > L ];
O[ not H[ neighbour ] ];
D[ G ];
H[ W * (N + 1) > L ].

*/
 
@interface SUAInclusiveTotal : HappySupplyAgent {
}

@end

/*

SUAExclusiveMean == Supply-Unhappy-Agent, exclusive mean variant

L = S / N;
A[ W > L ];
O[ not H[ neighbour ] ];
D[ G ];
H[ W > L ].

*/

@interface SUAExclusiveMean : ExtMeanHappySupplyAgent {
}

@end

/*

SUAExclusiveTotal == Supply-Unhappy-Agent, exclusive total variant

L = S;
A[ W * N > L ];
O[ not H[ neighbour ] ];
D[ G ];
H[ W * N > L ].

*/

@interface SUAExclusiveTotal : ExtAvgHappySupplyAgent {
}

@end
