/*
    CharityWorld: DoubleKnuthEpsilon.m
    Copyright (C) 2003  Macaulay Institute

    This file is part of CharityWorld, a simple agent-based model to
    demonstrate the effects of errors in floating point arithmetic on
    emergent outcomes.

    CharityWorld is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CharityWorld is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*

Implementation of DoubleKnuthEpsilon class

*/

#import "DoubleKnuthEpsilon.h"
#import <limits.h>
#import <math.h>

#ifdef INTEL
#  import <CIieeefp.h>
#endif

/*

max_exp(arg1, arg2)

Return a value equal to 2^x where x is the maximum exponent part of the
floating point arguments arg1 and arg2.

*/

static double max_exp(double arg1, double arg2) {
  int arg1exp, arg2exp;

  arg1exp = ilogb(arg1);
  if(arg1exp == INT_MAX) return arg1;
				// arg1 Inf or NaN
  arg2exp = ilogb(arg2);
  if(arg2exp == INT_MAX) return arg2;
				// arg2 Inf or NaN
  if(arg1exp == -INT_MAX && arg2exp == -INT_MAX) return 0.0;
				// arg1 and arg2 are both zero

  if(arg1exp > arg2exp) return scalbn(1.0, arg1exp);
  else return scalbn(1.0, arg2exp);
}

/*

deq(u, v, eps)

Implement Knuth's less stringent equality operator on p. 233, eq. (22)
(If we implement the more stringent one there will be cases where x != y 
and yet !(x < y) and !(x > y).)

*/

static BOOL deq(double u, double v, double eps) {
  BOOL result;
#ifdef INTEL
  fp_pctl pc = fpsetprecision(FP_PC_DBL);
#endif

  result = (fabs(v - u) <= eps * max_exp(u, v)) ? YES : NO;

#ifdef INTEL
  fpsetprecision(pc);
#endif
  return result;
}

/*

dne(u, v, eps)

Inequality is defined as the logical inverse of equality

*/

static BOOL dne(double u, double v, double eps) {
  return deq(u, v, eps) ? NO : YES;
}

/*

dgt(u, v, eps)

Implement Knuth's greater-than operator on p. 233, eq. (23)

*/

static BOOL dgt(double u, double v, double eps) {
  BOOL result;
#ifdef INTEL
  fp_pctl pc = fpsetprecision(FP_PC_DBL);
#endif

  result = (u - v > eps * max_exp(u, v)) ? YES : NO;

#ifdef INTEL
  fpsetprecision(pc);
#endif
  return result;
}
 
/*

dle(u, v, eps)

Define less-than-or-equal-to as the logical inverse of greater-than.

*/

static BOOL dle(double u, double v, double eps) {
  return dgt(u, v, eps) ? NO : YES;
}

/*

dlt(u, v, eps)

Implement Knuth's less-than operator on p. 233, eq. (21)

*/
 
static BOOL dlt(double u, double v, double eps) {
  BOOL result;
#ifdef INTEL
  fp_pctl pc = fpsetprecision(FP_PC_DBL);
#endif

  result = (v - u > eps * max_exp(u, v)) ? YES : NO;

#ifdef INTEL
  fpsetprecision(pc);
#endif
  return result;
}
  
/*

dge(u, v, eps)

Define greater-than-or-equal-to as the logical inverse of less-than.

*/

static BOOL dge(double u, double v, double eps) {
  return dlt(u, v, eps) ? NO : YES;
}



@implementation DoubleKnuthEpsilon

/* Methods to override from DoubleEpsilon */


-(BOOL)eq: (DoubleSimple *)value {
  return deq(d, [value get], [DoubleEpsilon getEpsilon]);
}

-(BOOL)ne: (DoubleSimple *)value {
  return dne(d, [value get], [DoubleEpsilon getEpsilon]);
}

-(BOOL)gt: (DoubleSimple *)value {
  return dgt(d, [value get], [DoubleEpsilon getEpsilon]);
}

-(BOOL)le: (DoubleSimple *)value {
  return dle(d, [value get], [DoubleEpsilon getEpsilon]);
}

-(BOOL)ge: (DoubleSimple *)value {
  return dge(d, [value get], [DoubleEpsilon getEpsilon]);
}

-(BOOL)lt: (DoubleSimple *)value {
  return dlt(d, [value get], [DoubleEpsilon getEpsilon]);
}


-(BOOL)ieq: (int)value {
  return deq(d, (double)value, [DoubleEpsilon getEpsilon]);
}

-(BOOL)ine: (int)value {
  return dne(d, (double)value, [DoubleEpsilon getEpsilon]);
}

-(BOOL)igt: (int)value {
  return dgt(d, (double)value, [DoubleEpsilon getEpsilon]);
}

-(BOOL)ile: (int)value {
  return dle(d, (double)value, [DoubleEpsilon getEpsilon]);
}

-(BOOL)ige: (int)value {
  return dge(d, (double)value, [DoubleEpsilon getEpsilon]);
}

-(BOOL)ilt: (int)value {
  return dlt(d, (double)value, [DoubleEpsilon getEpsilon]);
}


-(BOOL)ceq: (double)value {
  return deq(d, value, [DoubleEpsilon getEpsilon]);
}

-(BOOL)cne: (double)value {
  return dne(d, value, [DoubleEpsilon getEpsilon]);
}

-(BOOL)cgt: (double)value {
  return dgt(d, value, [DoubleEpsilon getEpsilon]);
}

-(BOOL)cle: (double)value {
  return dle(d, value, [DoubleEpsilon getEpsilon]);
}

-(BOOL)cge: (double)value {
  return dge(d, value, [DoubleEpsilon getEpsilon]);
}

-(BOOL)clt: (double)value {
  return dlt(d, value, [DoubleEpsilon getEpsilon]);
}


@end
