/*
    CharityWorld: DoubleShort.m
    Copyright (C) 2003  Macaulay Institute

    This file is part of CharityWorld, a simple agent-based model to
    demonstrate the effects of errors in floating point arithmetic on
    emergent outcomes.

    CharityWorld is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    CharityWorld is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*

Implementation for the DoubleShort class.

*/

#import "DoubleShort.h"

#import <stdlib.h>
#import <math.h>
#import <float.h>
#import <errno.h>
#import <string.h>
#import <stdlib.h>
#import <objc/objc-api.h>
#import <ctype.h>
#ifdef INTEL
#  import <CIieeefp.h>
static fp_pctl save_pctl;
#endif

#define NZ_ACC 40		// Number of zeros back from the 'e' in a
				// floating point number printed using %.*e
				// format required to be sure that the
				// floating point number is printed to
				// maximum accuracy

@implementation DoubleShort

+newUp {
  DoubleShort *obj = [super newUp];

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  obj->f = (float)obj->d;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return obj;
}

-copy {
  DoubleShort *obj = [super copy];

  [obj scset: f];

  return obj;
}

-set: (DoubleSimple *)value {
  [super set: value];
  if([value isKindOf: [DoubleShort class]]) {
    f = [(DoubleShort *)value sget];
  }
  else {
#ifdef INTEL
    save_pctl = fpsetprecision(FP_PC_SGL);
#endif

    f = (float)d;

#ifdef INTEL
    fpsetprecision(save_pctl);
#endif
  }
  return self;
}

-cset: (double)value {
  [super cset: value];

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  f = (float)d;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-scset: (float)value {
  f = value;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-setStr: (char *)str {
  f = [self scheckConst: str];

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-(double)get {
  double dd;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_DBL);
#endif

  dd = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return dd;
}

-(float)sget {
  return f;
}


-read: (FILE *)fp {

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  fscanf(fp, "%f", &f);

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-write: (FILE *)fp {

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  fprintf(fp, "%g", f);

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-write: (FILE *)fp sf: (int)sigfig {

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  fprintf(fp, "%.*e", sigfig - 1, f);

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-writeAccurately: (FILE *)fp {
  char *buf, tbuf[3];
  int dp;
  int nz;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  dp = NZ_ACC;
  do {
    char *p, *q;
    int len = snprintf(tbuf, 2, "%.*e", dp, f);
				// Use snprintf and a dummy buffer to get the
				// size of buffer needed to print the floating
				// point number to the required number of
				// decimal places
    if(len < 0) {
      perror("Error in call to snprintf");
      abort();
    }

    buf = (char *)malloc(len + 1);
				// Allocate memory for the buffer
    if(buf == NULL) {
      perror("Memory allocation");
      abort();
    }

    sprintf(buf, "%.*e", dp, f);
				// Print the floating point number to the
				// buffer in the format X.YYYYYY0000eZZZ
    p = strchr(buf, (int)'e');
				// Find the 'e' in the buffer.
    if(p == NULL) {		// If we can't find the 'e' then this is some
				// funny number like NaN or Inf
      fprintf(fp, "%e", f);
      free(buf);

#ifdef INTEL
      fpsetprecision(save_pctl);
#endif

      return self;
    }
    q = p;			// Remember where it is...
      
    p--;			// Look back from the 'e' 
    nz = 0;			// and count the number
    while((*p) == '0') {	// of consecutive '0's
      nz++;
      p--;
    }
				// Right now, this should be the state of play:
				// X.YYYYYY0000eZZZ
				// ^buf   ^p   ^q
				// If the number is like X.0000000000eZZZ,
				// then p points here:    ^p.

    if(nz >= NZ_ACC) {		// We've got enough zeros before the 'e' to
				// ensure that maximum accuracy has been
				// achieved (according to the NZ_ACC parameter)
      if((*p) != '.') {
	p++;			// p didn't point to the decimal place, so
				// shift it right one place to point to the
				// first in a long line of zeros.
      }
      (*p) = '\0';		// Overwrite the decimal point or the first
				// in a long line of zeros with a string
				// terminator. Now we have (using # for the
				// terminator):
				// X.YYYYYY#000eZZZ# or X#0000000000eZZZ#
				// ^buf    ^p  ^q       ^buf        ^q
      fprintf(fp, "%s%s", buf, q);
				// So, printing buf then q gives:
				// X.YYYYYYeZZZ      or XeZZZ
    }
    else {
      dp += NZ_ACC;		// Increment the number of decimal places to
				// print by the required accuracy
    }

    free(buf);			// Free the buffer
  } while(nz < NZ_ACC);

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}


-abs {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  if(f < 0.0F) f *= -1.0F;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}


-add: (DoubleSimple *)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  if([value isKindOf: [DoubleShort class]]) {
    f += [(DoubleShort *)value sget];
  }
  else {
    f += (float)[value get];
  }

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-sub: (DoubleSimple *)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  if([value isKindOf: [DoubleShort class]]) {
    f -= [(DoubleShort *)value sget];
  }
  else {
    f -= (float)[value get];
  }

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-mul: (DoubleSimple *)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  if([value isKindOf: [DoubleShort class]]) {
    f *= [(DoubleShort *)value sget];
  }
  else {
    f *= (float)[value get];
  }

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-div: (DoubleSimple *)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  if([value isKindOf: [DoubleShort class]]) {
    f /= [(DoubleShort *)value sget];
  }
  else {
    f /= (float)[value get];
  }

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}


-iadd: (int)value {
#ifdef INTEL
    save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  f += (float)value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-isub: (int)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  f -= (float)value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-imul: (int)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  f *= (float)value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-idiv: (int)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  f /= (float)value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}


-cadd: (double)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  f += (float)value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-csub: (double)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  f -= (float)value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-cmul: (double)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  f *= (float)value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-cdiv: (double)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  f /= (float)value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}


-scadd: (float)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  f += value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-scsub: (float)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  f -= value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-scmul: (float)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  f *= value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}

-scdiv: (float)value {
#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  f /= value;

#ifdef INTEL
  fpsetprecision(FP_PC_DBL);
#endif

  d = (double)f;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return self;
}


-(BOOL)eq: (DoubleSimple *)value {
  if([value isKindOf: [DoubleShort class]]) {
    return f == [(DoubleShort *)value sget] ? YES : NO;
  }
  else {
    BOOL result;

#ifdef INTEL
    save_pctl = fpsetprecision(FP_PC_SGL);
#endif

    result = (double)f == [value get] ? YES : NO;

#ifdef INTEL
    fpsetprecision(save_pctl);
#endif

    return result;
  }
}

-(BOOL)ne: (DoubleSimple *)value {
  if([value isKindOf: [DoubleShort class]]) {
    return f != [(DoubleShort *)value sget] ? YES : NO;
  }
  else {
    BOOL result;

#ifdef INTEL
    save_pctl = fpsetprecision(FP_PC_SGL);
#endif

    result = (double)f != [value get] ? YES : NO;

#ifdef INTEL
    fpsetprecision(save_pctl);
#endif

    return result;
  }
}

-(BOOL)gt: (DoubleSimple *)value {
  if([value isKindOf: [DoubleShort class]]) {
    return f > [(DoubleShort *)value sget] ? YES : NO;
  }
  else {
    BOOL result;

#ifdef INTEL
    save_pctl = fpsetprecision(FP_PC_SGL);
#endif

    result = (double)f > [value get] ? YES : NO;

#ifdef INTEL
    fpsetprecision(save_pctl);
#endif

    return result;
  }
}

-(BOOL)le: (DoubleSimple *)value {
  if([value isKindOf: [DoubleShort class]]) {
    return f <= [(DoubleShort *)value sget] ? YES : NO;
  }
  else {
    BOOL result;

#ifdef INTEL
    save_pctl = fpsetprecision(FP_PC_SGL);
#endif

    result = (double)f <= [value get] ? YES : NO;

#ifdef INTEL
    fpsetprecision(save_pctl);
#endif

    return result;
  }
}

-(BOOL)ge: (DoubleSimple *)value {
  if([value isKindOf: [DoubleShort class]]) {
    return f >= [(DoubleShort *)value sget] ? YES : NO;
  }
  else {
    BOOL result;

#ifdef INTEL
    save_pctl = fpsetprecision(FP_PC_SGL);
#endif

    result = (double)f >= [value get] ? YES : NO;

#ifdef INTEL
    fpsetprecision(save_pctl);
#endif

    return result;
  }
}

-(BOOL)lt: (DoubleSimple *)value {
  if([value isKindOf: [DoubleShort class]]) {
    return f < [(DoubleShort *)value sget] ? YES : NO;
  }
  else {
    BOOL result;

#ifdef INTEL
    save_pctl = fpsetprecision(FP_PC_SGL);
#endif

    result = (double)f < [value get] ? YES : NO;

#ifdef INTEL
    fpsetprecision(save_pctl);
#endif

    return result;
  }
}


-(BOOL)ieq: (int)value {
  BOOL result;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  result = f == (float)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)ine: (int)value {
  BOOL result;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  result = f != (float)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)igt: (int)value {
  BOOL result;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  result = f > (float)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)ile: (int)value {
  BOOL result;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  result = f <= (float)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)ige: (int)value {
  BOOL result;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  result = f >= (float)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)ilt: (int)value {
  BOOL result;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  result = f < (float)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}


-(BOOL)ceq: (double)value {
  BOOL result;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  result = f == (float)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)cne: (double)value {
  BOOL result;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  result = f != (float)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)cgt: (double)value {
  BOOL result;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  result = f > (float)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)cle: (double)value {
  BOOL result;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  result = f <= (float)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)cge: (double)value {
  BOOL result;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  result = f >= (float)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}

-(BOOL)clt: (double)value {
  BOOL result;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  result = f < (float)value ? YES : NO;

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return result;
}


-(BOOL)sceq: (float)value {
  return f == value ? YES : NO;
}

-(BOOL)scne: (float)value {
  return f != value ? YES : NO;
}

-(BOOL)scgt: (float)value {
  return f > value ? YES : NO;
}

-(BOOL)scle: (float)value {
  return f <= value ? YES : NO;
}

-(BOOL)scge: (float)value {
  return f >= value ? YES : NO;
}

-(BOOL)sclt: (float)value {
  return f < value ? YES : NO;
}


-(float)scheckConst: (const char *)value {
  float tmp;

#ifdef INTEL
  save_pctl = fpsetprecision(FP_PC_SGL);
#endif

  sscanf(value, "%f", &tmp);

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return tmp;
}

@end
